#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Cube multidimensionnel - énoncé
# 
# Ce notebook aborde différentes solutions pour traiter les données qu'on représente plus volontiers en plusieurs dimensions. Le mot-clé associé est [OLAP](https://en.wikipedia.org/wiki/Online_analytical_processing) ou cube OLAP. [Mondrian](http://community.pentaho.com/projects/mondrian/) est une solution open source, [cubes](http://cubes.databrewery.org/) est écrit en python. [Kylin](http://kylin.apache.org/) propose ce service sur des données stockées sur Hadoop. L'objectif est ici de découvrir pas d'explorer ces solutions.

# In[1]:


get_ipython().run_line_magic('matplotlib', 'inline')
import matplotlib.pyplot as plt
plt.style.use('ggplot')
import pyensae
from pyquickhelper.helpgen import NbImage
from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ### Représentation

# Le module [pandas](http://pandas.pydata.org/) manipule des tables et c'est la façon la plus commune de représenter les données. Lorsque les données sont multidimensionnelles, on distingue les coordonnées des valeurs :

# In[2]:


NbImage("cube1.png")


# Dans cet exemple, il y a :
#     
# * 3 coordonnées : Age, Profession, Annéee
# * 2 valeurs : Espérance de vie, Population
# 
# On peut représenter les donnés également comme ceci :

# In[3]:


NbImage("cube2.png")


# C'est assez simple. Prenons un exemple : [table de mortalité de 1960 à 2010](http://www.data-publica.com/opendata/7098--population-et-conditions-sociales-table-de-mortalite-de-1960-a-2010) qu'on récupère à l'aide de la fonction [table_mortalite_euro_stat](http://www.xavierdupre.fr/app/actuariat_python/helpsphinx/actuariat_python/data/population.html#actuariat_python.data.population.table_mortalite_euro_stat). C'est assez long (4-5 minutes) sur l'ensemble des données car elles doivent être prétraitées (voir la documentation de la fonction). Pour écouter, il faut utiliser le paramètre *stop_at*.

# In[4]:


from actuariat_python.data import table_mortalite_euro_stat 
table_mortalite_euro_stat()


# In[5]:


import os
os.stat("mortalite.txt")


# In[6]:


import pandas
df = pandas.read_csv("mortalite.txt", sep="\t", encoding="utf8", low_memory=False)
df.head()


# ### Cubes de données avec xarray

# #### création
# 
# Cette réprésentation sous forme de table n'est pas toujours très lisible. Les informations sont souvent répétées et les données sont vraiment multidimensionnelles. Le module [xarray](http://xarray.pydata.org/en/stable/) propose de distinguer les coordonnées des valeurs pour proposer des manipulations plus intuitives. Le module propose un DataFrame multidimensionnelle [DataSet](http://xarray.pydata.org/en/stable/data-structures.html#dataset).

# In[7]:


from actuariat_python.data import table_mortalite_euro_stat 
table_mortalite_euro_stat()
import pandas
df = pandas.read_csv("mortalite.txt", sep="\t", encoding="utf8", low_memory=False)
df.columns


# Dans ce cas-ci, pour reprendre la temrinologie du module [xarray](http://xarray.pydata.org/en/stable/), nous avons :
# 
# * les **dimensions** : annee, age_num, pays, genre
# * les **valeurs** : une valeur par indicateur
# 
# On peut passer d'un [DataFrame](http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.html) à un [DataSet](http://xarray.pydata.org/en/stable/data-structures.html#dataset) de la façon suivante :
# 
# * les colonnes indéxées représentent les dimensions
# * les colonnes non indéxées sont les valeurs
# 
# On garde supprime les colonnes qui ne nous intéressent pas et les valeurs manquantes :

# In[8]:


df2 = df[["annee", "age_num","indicateur","pays","genre","valeur"]].dropna().reset_index(drop=True)
df.columns, df2.columns


# In[9]:


df2.head()


# In[10]:


df2["indicateur"] = df2["indicateur"].astype(str)
df2["genre"] = df2["genre"].astype(str)
df2["pays"] = df2["pays"].astype(str)


# In[11]:


df2.dtypes


# L'opération complexe consiste à faire passer les valeurs de la colonne *indicateur* en tant que colonnes. C'est l'objet de la méthode [pivot_table](http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.pivot_table.html?highlight=pivot#pandas.DataFrame.pivot_table) :

# In[12]:


piv = df2.pivot_table(index=["annee", "age_num","pays","genre"],
                columns=["indicateur"],
               values="valeur")
piv.head()


# In[13]:


piv.dtypes


# Les données sont maintenant prêtes à passer sous *xarray* :

# In[14]:


import xarray
ds = xarray.Dataset.from_dataframe(piv)
ds


# #### sélection
# 
# Il est facile ensuite d'extraire les données d'un pays avec la méthode [sel](http://xarray.pydata.org/en/stable/indexing.html#indexing-and-selecting-data) :

# In[15]:


ds.sel(pays=["FR"])


# Ou de plusieurs pays :

# In[16]:


ds.sel(pays=["FR", "BE"])


# Ou plusieurs dimensions :

# In[17]:


ds.sel(pays="FR", annee=2000)


# Pour accéder à la série ``LIFEXP`` pour les hommes, les années 2000 et 2010, le pays ``FR``, on fait la différence, puis on la récupère sous forme de DataFrame :

# In[18]:


(ds.sel(pays="FR", annee=2010, genre="T")["LIFEXP"] - 
 ds.sel(pays="FR", annee=2000, genre="T")["LIFEXP"]).to_dataframe().head()


# On a gagné presque deux ans et demi d'espérance de vie à la naissance en 10 ans.

# In[19]:


(ds.sel(pays="FR", annee=2010, genre=["F","M"])["LIFEXP"] - 
 ds.sel(pays="UK", annee=2010, genre=["F","M"])["LIFEXP"]).to_dataframe().head()


# #### Quelques graphes

# In[20]:


ds.sel(annee=2010,age_num=1,genre="T")["LIFEXP"].to_dataframe() \
            .sort_values("LIFEXP", ascending=False) \
            .plot(y="LIFEXP", kind="bar", figsize=(16,6), ylim=[65,85])


# In[21]:


ds.sel(annee=2010,genre="T",pays="FR")["LIFEXP"].plot()


# ### Exercice 1 : que font les lignes suivantes ?
# 
# On pourra s'aider des pages :
# 
# * [align nad reindex](http://xarray.pydata.org/en/stable/indexing.html#align-and-reindex)
# * [transforming datasets](http://xarray.pydata.org/en/stable/data-structures.html?highlight=assign#transforming-datasets)

# In[22]:


ds.assign(LIFEEXP_add = ds.LIFEXP-1)


# In[23]:


meanp = ds.mean(dim="pays")
ds1, ds2 = xarray.align(ds, meanp, join='outer')


# In[24]:


joined = ds1.assign(meanp = ds2["LIFEXP"])


# In[25]:


joined.to_dataframe().head()


# ### Lire, écrire des datasets

# Le module [xarray](http://xarray.pydata.org/en/stable/) s'appuie sur le module [netCDF4](https://github.com/Unidata/netcdf4-python) qui lui-même est un [wrapper](https://fr.wikipedia.org/wiki/Adaptateur_(patron_de_conception)) poython de la libraire [netCDF-c](https://github.com/Unidata/netcdf-c). Celle-ci est spécialisée dans la lecture et l'écriture de données scientifiques. Concrètement, ce n'est pas [xarray](http://xarray.pydata.org/en/stable/) qui s'en occupe mais [netCDF4](https://github.com/Unidata/netcdf4-python). Cela explique la syntaxe décrite par [Serialization and IO](http://xarray.pydata.org/en/stable/io.html#serialization-and-io) :

# In[26]:


ds.data_vars


# In[27]:


try:
    ds.to_netcdf('mortalite.nc')
except ValueError as e:
    # it breaks with pandas 0.17
    # xarray has to be updated
    print("l'écriture a échoué")
    pass


# C'est un format binaire plus efficace que le format texte :

# In[28]:


import os
if os.path.exists("mortalite.nc"):
    os.stat('mortalite.nc').st_size, os.stat('mortalite.txt').st_size


# On relit :

# In[29]:


if os.path.exists("mortalite.nc"):
    ds_lu = xarray.open_dataset('mortalite.nc')
    ds_lu


# Le module [xarray](http://xarray.pydata.org/en/stable/) propose également de lire des données de plusieurs fichiers pour ne former qu'un seul dataset (voir [Combining multiple files](http://xarray.pydata.org/en/stable/io.html#combining-multiple-files)) :

# In[30]:


pays = list(_.values for _ in ds["pays"])
pays[:5]


# In[31]:


for p in pays[:5]:
    print("enregistre", p)
    d = ds.sel(pays=[p])
    try:
        d.to_netcdf("mortalite_pays_%s.nc" % p)
    except ValueError:
        print("l'écriture a échoué pour", p)


# On relit :

# In[32]:


import os
if os.path.exists("mortalite_pays_AM.nc"):
    ds_lu2 = xarray.open_mfdataset('mortalite_pays*.nc')
    ds_lu2


# A quoi ça sert ?
# 
# Lorsqu'on travaille avec des gros fichiers de données, on cherche à retarder le plus possible l'exécution d'un calcul, on manipule les données sans les charger en mémoire. On appelle ceci [évaluation paresseuse](https://fr.wikipedia.org/wiki/%C3%89valuation_paresseuse) ou [lazy evaluation](https://en.wikipedia.org/wiki/Lazy_evaluation) en anglais. [xarray](http://xarray.pydata.org/en/stable/) et [dask](http://dask.pydata.org/en/latest/) permettent de faire cela (voir [Using dask with xarray](http://xarray.pydata.org/en/stable/dask.html#dask-io)). La logique qu'on suit est alors assez proche de la [programmation fonctionnelle](https://fr.wikipedia.org/wiki/Programmation_fonctionnelle).

# In[33]: